import pygame
from twisted.internet import reactor
from twisted.internet.protocol import Protocol, ClientFactory
from struct import *
import random, os, sys

import gobject
gobject.threads_init()
import pygst
pygst.require ("0.10")
import gst



SCREEN_WIDTH = 640
SCREEN_HEIGHT = 480
currentTotal = 0
currentSamples = 0
images = {}
sound = {}
alienSpeed = 0.6
speedChanged = False
score = 0
lives = 300
level = 1
newLevel = False
invaded = False
dir = 0.0

bullets = pygame.sprite.RenderClear()

# loadImageFile from Kevin Harris's pygame tutorial
#

def loadImageFile( name, useColorKey = False ):

    fullname = os.path.join( "data", name )

    try:
        image = pygame.image.load( fullname )
    except pygame.error, message:
        print "Cannot load image:", fullname
        raise SystemExit, message

    # To speed things up, convert the images to SDL's internal format.
    image = image.convert()

    if useColorKey is True:
        # Use the color of the pixel located at (0,0) as our colorkey
        colorkey = image.get_at( (0,0) )
        image.set_colorkey( colorkey, pygame.RLEACCEL )

    # Return the image
    return image

def loadSoundFile(name):
	fullname = os.path.join("data",name)
	
	try:
		sound = pygame.mixer.Sound(fullname)
	except pygame.error, message:
		print "Cannot load sound;", fullname
		raise SystemExit, message

	return sound

def handleSlice(slice):
	global currentTotal, currentSamples
	for i in range(N):
		for j in range(M):
			for k in range(O):
				offset = i*M*O*4 + j*O*4 + k*4
				subdata = slice[offset:offset+4]
				(val,) = unpack('f', subdata)
				currentTotal += val
				currentSamples += 1

def initialiseImages():
	global images
	images = { 'background': loadImageFile("background.PNG"),
				'enemy1': loadImageFile("enemy1.PNG", True),
				'enemy2': loadImageFile("enemy2.png", True),
				'enemy3': loadImageFile("enemy3.PNG", True),
				'player': loadImageFile("spaceship.PNG", True),
				'bullet': loadImageFile("bullet.png", True),
				'explosion': loadImageFile("explosion1.bmp", True) }	

def initialiseSound():
	global sound
	sound = {'explode': loadSoundFile("invaderkilled.wav"),
			'shoot': loadSoundFile("shoot.wav"),
			'spawn': loadSoundFile("Respawn.wav")
	}

class Alien (pygame.sprite.Sprite):
	def __init__(self, image, position):
		pygame.sprite.Sprite.__init__( self )
		self.image = image
		self.rect = self.image.get_rect()
		self.rect.midtop = position
		self.moveLeft = True
		self.moves = 4

	def update(self):
		if newLevel:
			self.moves = 4
			self.moveLeft = True
		if self.moves == 8:
			self.moves = 0
			self.moveLeft = not self.moveLeft
			self.rect.move_ip(0,10)
			global alienSpeed, speedChanged
			if not speedChanged:
				alienSpeed *= 0.9
				speedChanged = True

		elif self.moveLeft:
			self.rect.move_ip(-15,0)
			self.moves += 1
		else:
			self.rect.move_ip(15,0) 
			self.moves += 1

		if self.rect.bottom >= SCREEN_HEIGHT-50:
			global invaded
			invaded = True

class Player (pygame.sprite.Sprite):
	def __init__(self):
		pygame.sprite.Sprite.__init__(self)
		self.image = images['player']
		self.rect = self.image.get_rect()
		self.rect.center = (SCREEN_WIDTH/2, SCREEN_HEIGHT - 60)
		self.shot_counter = 0
			
	def update(self):
        # avg from -100.0 to 100.0
		global dir
		avg = dir

		max_speed = 20
		global currentTotal, currentSamples
		#if currentSamples == 0:
		#	return
		#avg = currentTotal / currentSamples
		currentTotal = 0
		currentSamples = 0
		self.rect.move_ip(max_speed*avg / 100.0, 0)
		if self.rect.right > 640:
			self.rect.right = 640
		if self.rect.left < 0:
			self.rect.left = 0

		if self.shot_counter == 4:
			bullets.add(Bullet(self.rect.midtop))
			self.shot_counter = 0
		else:
			self.shot_counter += 1

class Bullet(pygame.sprite.Sprite):
	def __init__(self, pos):
		pygame.sprite.Sprite.__init__( self )
		self.image = images['bullet']
		self.rect = self.image.get_rect()
		self.rect.midbottom = pos
		sound['shoot'].play()
	
	def update(self):
		self.rect.top -= 9
		if self.rect.top < 0:
			self.kill()		
	

def initialiseAliens():
	aliens = pygame.sprite.RenderClear()
	for r in range(5):
		for c in range (9):
			if r == 0:
				image = images['enemy1']
			elif r < 3:
				image = images['enemy2']
			else:
				image = images['enemy3']

			x = 90 + c * 58
			y = 60 + r * 48

			aliens.add( Alien(image, (x,y)))

	sound['spawn'].play()
	return aliens

class GstringClientProtocol(Protocol):
	def __init__(self):
		self.buffer = ""

	def connectionMade(self):
		self.transport.write(pack('c', chr(42)))
		self.new = 1

	def dataReceived(self, data):
		if self.new:
			global N, M, O
			info = data[:3]
			data = data[3:]
			(N,M,O,) = unpack('bbb', info)
			self.new = 0 
			return
		self.buffer += data
		while len(self.buffer) > N*M*O*4:
			slice = self.buffer[:(N*M*O*4)]
			self.buffer = self.buffer[(N*M*O*4):]
			handleSlice(slice)

class GstringClientFactory(ClientFactory):
	protocol = GstringClientProtocol

def updateAliens():
	global aliens, newLevel
	aliens.update()
	aliens.clear(screen, images['background'])
	aliens.draw(screen)
	global speedChanged
	speedChanged = False
	newLevel = False
	if len(aliens) == 0:
		global level, alienSpeed, bullets
		level+=1
		alienSpeed = 0.6 - (level-1)*0.05 
		aliens = initialiseAliens()
		bullets.empty()
		newLevel = True
	reactor.callLater(alienSpeed, updateAliens)
		
def updatePlayer():
	global invaded
	if invaded:
		invaded = False
		global aliens,bullets,lives, level,alienSpeed
		aliens.clear(screen, images['background'])
		bullets.clear(screen, images['background'])
		aliens.empty()
		bullets.empty()
		aliens = initialiseAliens()
		lives -= 1
		alienSpeed = 0.6 - (level-1)*0.05
		if lives == 0:	
			gameOver()
	else:
		player.update()
		player.clear(screen, images['background'])
		player.draw(screen)
	reactor.callLater(0.1, updatePlayer)

def updateBullets():
	bullets.update()
	bullets.clear(screen, images['background'])
	bullets.draw(screen)
	reactor.callLater(0.05, updateBullets)

def collisionDetection():
	reactor.callLater(0.1, collisionDetection)
	for x in pygame.sprite.groupcollide( aliens, bullets, 1, 1):
			global score, sound
			score += 10* level
			sound['explode'].play()

def flip():
	pygame.display.flip()
	reactor.callLater(0.05, flip)

def updateScore():
	text = gameFont.render('Space Invaders      LEVEL: ' + str(level), True, (255,255, 255), (0,0,0))
	textRect = text.get_rect()
	textRect.topleft = (20,20)
	screen.blit(text, textRect)
	score_text = gameFont.render('SCORE: %10s       LIVES: %s' % (score,lives), True, (255,255,255), (0,0,0))
	score_rect = score_text.get_rect()
	score_rect.topleft = (400, 20)
	screen.blit(score_text, score_rect)
	reactor.callLater(0.2, updateScore)

def gameOver():
	sleep(100)
	reactor.stop()
	pygame.quit()

def updateSpeedCb(el, param):
    global dir
    dir = el.props.direction
    player.update()
        

random.seed()
pygame.init()
gameFont = pygame.font.Font(None,24)
screen = pygame.display.set_mode( (SCREEN_WIDTH, SCREEN_HEIGHT) )
pygame.display.set_caption( "Space Invaders" )
initialiseImages()
initialiseSound()
screen.blit(images['background'], (0,0))

aliens = initialiseAliens()
player = pygame.sprite.RenderClear()
player.add(Player())
sound['spawn'].play()

pipeline = gst.parse_launch("v4lsrc autoprobe-fps=false ! ffmpegcolorspace ! redgreentv name=rg red-weight=160.0 green-weight=90.0 ! ffmpegcolorspace ! xvimagesink")
rg = pipeline.get_by_name('rg')
rg.connect('notify::direction', updateSpeedCb)
print pipeline.set_state(gst.STATE_PLAYING)

#host = sys.argv[1]
#port = int(sys.argv[2])
#reactor.connectTCP(host, port, GstringClientFactory())
reactor.callLater(0.5, updateAliens)
reactor.callLater(0.1, updatePlayer)
reactor.callLater(0.05, updateBullets)
reactor.callLater(0.2, collisionDetection)
reactor.callLater(0.05, flip)
reactor.callLater(0.2, updateScore)
reactor.run()
